module basic_functions
  implicit none
contains
  function break_next_fibre_els(N,fibres,thresh,broken) result(x)
    integer, intent(in) :: N
    integer, dimension(N,N,N,N,N) :: fibres
    real*8, dimension(N,N,N,N,N), intent(in) :: thresh
    integer :: broken
    real*8 :: x  !Extension at which the fibre breaks
    !Internal variables
    real*8 :: low
    logical :: found
    integer :: i, j, k, l, m
    integer :: ipos, jpos, kpos, lpos, mpos
    
    !Find the next fibre to break
    found = .false.
    low = 0d0
    ipos = 0
    jpos = 0
    kpos = 0
    lpos = 0
    mpos = 0
    do m = 1,N
    do l = 1,N
       do k = 1,N
          do j = 1,N
             do i = 1,N
                if (.not. found) then
                   if (fibres(i,j,k,l,m) == 1) then
                      low = thresh(i,j,k,l,m)
                      ipos = i
                      jpos = j
                      kpos = k
                      lpos = l
                      mpos = m
                      found = .true.
                   endif
                else
                   if (fibres(i,j,k,l,m) == 1 .and. thresh(i,j,k,l,m) < low) then
                      low = thresh(i,j,k,l,m)
                      ipos = i
                      jpos = j
                      kpos = k
                      lpos = l
                      mpos = m
                   endif
                endif
             enddo
          enddo
       enddo
    enddo
    enddo
    
    !Break the next fibre
    fibres(ipos,jpos,kpos,lpos,mpos) = 0

    !Return the threshold at which the fibre is broken
    x = thresh(ipos,jpos,kpos,lpos,mpos)/(1d0+DBLE(broken)/DBLE(N**5-broken))
    !Update number of broken fibres
    broken = broken + 1
  end function break_next_fibre_els

!Break the next fibre in local load sharing
  function break_next_fibre_lls(N,fibres,thresh,hoshk,holes) result(x)
    integer, intent(in) :: N
    integer, dimension(N,N,N,N,N) :: fibres
    real*8, dimension(N,N,N,N,N,2) :: thresh  !Thresholds for each fibre
    integer, dimension(N**5,3) :: hoshk
    integer, dimension(N,N,N,N,N,10) :: holes
    real*8 :: x
    !Internal variables
    real*8 :: thresh_factor, low
    integer :: i, j, k, l, m, ijklm, ipos, jpos, kpos, lpos, mpos, a
    logical :: found
    
    !Update effective thresholds for each fibre
    do m = 1,N
    do l = 1,N
    do k = 1,N
       do j = 1,N
          do i = 1,N
             thresh_factor = 1d0
             do a = 1,4
                if (holes(i,j,k,l,m,a) .ne. 0) then
                   ijklm = holes(i,j,k,l,m,a)
                   thresh_factor = thresh_factor + DBLE(hoshk(ijklm,2))/DBLE(hoshk(ijklm,3))
                endif
             enddo
             thresh(i,j,k,l,m,2) = thresh(i,j,k,l,m,1)/thresh_factor
          enddo
       enddo
    enddo
    enddo
    enddo
    
    !Find the next fibre to break
    found = .false.
    low = 0d0
    ipos = 0
    jpos = 0
    kpos = 0
    lpos = 0
    mpos = 0
    do m = 1,N
    do l = 1,N
       do k = 1,N
          do j = 1,N
             do i = 1,N
                if (.not. found) then
                   if (fibres(i,j,k,l,m) == 1) then
                      low = thresh(i,j,k,l,m,2)
                      ipos = i
                      jpos = j
                      kpos = k
                      lpos = l
                      mpos = m
                      found = .true.
                   endif
                else
                   if (fibres(i,j,k,l,m) == 1 .and. thresh(i,j,k,l,m,2) < low) then
                      low = thresh(i,j,k,l,m,2)
                      ipos = i
                      jpos = j
                      kpos = k
                      lpos = l
                      mpos = m
                   endif
                endif
             enddo
          enddo
       enddo
    enddo
    enddo
    
    !Break the next fibre
    fibres(ipos,jpos,kpos,lpos,mpos) = 0

    !Return x-value at which the fibre breaks
    x = thresh(ipos,jpos,kpos,lpos,mpos,2)
  end function break_next_fibre_lls
  
!The Hoshen-Kopelman algorithm to find holes + the neighbours
  subroutine hoshen_kopelman(N,fibres,hoshk,holes)
    integer, intent(in) :: N  !Dimension of grid
    integer, dimension(N,N,N,N,N), intent(in) :: fibres
    integer, dimension(N**5,3), intent(out) :: hoshk !Equivalence relation array
    integer, dimension(N,N,N,N,N,10), intent(out) :: holes  !Neighbouring holes
    !Internal variables
    integer :: i, j, k, l, m, ijklm, a  !Iteration variables (array positions)
    integer :: right1, left1, right2, left2, right3, left3
    integer :: right4, left4, right5, left5  !Positions of neighbouring fibres
    integer :: inext, iprev, jnext, jprev, knext, kprev
    integer :: lnext, lprev, mnext, mprev !Positions of neighbouring fibres
    integer :: Nf, N1, N2, N3, N4, N5
    logical :: already_found
    integer :: counter
    
    hoshk = 0
    do ijklm = 1,N**5
       hoshk(ijklm,1) = ijklm
    enddo
    
    do m = 1,N
    do l = 1,N
       do k = 1,N
          do j = 1,N
             do i = 1,N
                if (fibres(i,j,k,l,m) == 0) then
                   !Initialize positions of the fibres to the "right"
                   ijklm = i + (j-1)*N + (k-1)*N**2 + (l-1)*N**3 + (m-1)*N**4
                   !Dimension 1
                   if (i == N) then
                      right1 = ijklm - (N-1)
                      inext = 1
                   else
                      right1 = ijklm + 1
                      inext = i + 1
                   endif
                   !Dimension 2
                   if (j == N) then
                      right2 = ijklm - (N-1)*N
                      jnext = 1
                   else
                      right2 = ijklm + N
                      jnext = j + 1
                   endif
                   !Dimension 3
                   if (k == N) then
                      right3 = ijklm - (N-1)*N**2
                      knext = 1
                   else
                      right3 = ijklm + N**2
                      knext = k + 1
                   endif
                   !Dimension 4
                   if (l == N) then
                      right4 = ijklm - (N-1)*N**3
                      lnext = 1
                   else
                      right4 = ijklm + N**3
                      lnext = l + 1
                   endif
                   !Dimension 5
                   if (m == N) then
                      right5 = ijklm - (N-1)*N**4
                      mnext = 1
                   else
                      right5 = ijklm + N**4
                      mnext = m + 1
                   endif
                   
                   Nf = hoshk(ijklm,1)
                   do while (Nf .ne. hoshk(Nf,1))
                      Nf = hoshk(Nf,1)
                   enddo
                   
                   !If the one to the right in D1 is broken as well
                   if (fibres(inext,j,k,l,m) == 0) then
                      N1 = hoshk(right1,1)
                      do while (N1 .ne. hoshk(N1,1))
                         N1 = hoshk(N1,1)
                      enddo
                      if (Nf .ne. N1) then
                         hoshk(MAX(Nf,N1),1) = MIN(Nf,N1)
                      endif
                   endif
                   
                   !If the one to the right in D2 is broken as well
                   if (fibres(i,jnext,k,l,m) == 0) then
                      N2 = hoshk(right2,1)
                      do while (N2 .ne. hoshk(N2,1))
                         N2 = hoshk(N2,1)
                      enddo
                      if (Nf .ne. N2) then
                         hoshk(MAX(Nf,N2),1) = MIN(Nf,N2)
                      endif
                   endif
                   
                   !If the one to the right in D3 is broken as well
                   if (fibres(i,j,knext,l,m) == 0) then
                      N3 = hoshk(right3,1)
                      do while (N3 .ne. hoshk(N3,1))
                         N3 = hoshk(N3,1)
                      enddo
                      if (Nf .ne. N3) then
                         hoshk(MAX(Nf,N3),1) = MIN(Nf,N3)
                      endif
                   endif
                   
                   !If the one to the right in D4 is broken as well
                   if (fibres(i,j,k,lnext,m) == 0) then
                      N4 = hoshk(right4,1)
                      do while (N4 .ne. hoshk(N4,1))
                         N4 = hoshk(N4,1)
                      enddo
                      if (Nf .ne. N4) then
                         hoshk(MAX(Nf,N4),1) = MIN(Nf,N4)
                      endif
                   endif
                   
                   !If the one to the right in D5 is broken as well
                   if (fibres(i,j,k,l,mnext) == 0) then
                      N5 = hoshk(right5,1)
                      do while (N5 .ne. hoshk(N5,1))
                         N5 = hoshk(N5,1)
                      enddo
                      if (Nf .ne. N5) then
                         hoshk(MAX(Nf,N5),1) = MIN(Nf,N5)
                      endif
                   endif

                endif
             enddo
          enddo
       enddo
    enddo
    enddo

    !Clean up the hoshk array to give all equivalence classes the same values
    do ijklm = 1,N**5
       Nf = hoshk(ijklm,1)
       do while (Nf .ne. hoshk(Nf,1))
          Nf = hoshk(Nf,1)
       enddo
       hoshk(ijklm,1) = Nf
    enddo
    
    !Find the number of broken fibres in each hole and its number of neighbours
    holes = 0
    do m = 1,N
    do l = 1,N
       do k = 1,N
          do j = 1,N
             do i = 1,N
                ijklm = i + (j-1)*N + (k-1)*N**2 + (l-1)*N**3 + (m-1)*N**4
                
                if (fibres(i,j,k,l,m) == 0) then
                   !If this is broken, add one to the counter for fibres in this hole
                   hoshk(hoshk(ijklm,1),2) = hoshk(hoshk(ijklm,1),2) + 1
                else
                   counter = 0
                   !Find positions of neighbouring fibres:

                   !Dimension 1
                   if (i == N) then
                      right1 = ijklm - (N-1)
                      left1 = ijklm - 1
                      inext = 1
                      iprev = N - 1
                   else if (i == 1) then
                      right1 = ijklm + 1
                      left1 = ijklm + (N-1)
                      inext = 2
                      iprev = N
                   else
                      right1 = ijklm + 1
                      left1 = ijklm - 1
                      inext = i + 1
                      iprev = i - 1
                   endif
                   
                   !Dimension 2
                   if (j == N) then
                      right2 = ijklm - (N-1)*N
                      left2 = ijklm - N
                      jnext = 1
                      jprev = N - 1
                   else if (j == 1) then
                      right2 = ijklm + N
                      left2 = ijklm + (N-1)*N
                      jnext = 2
                      jprev = N
                   else
                      right2 = ijklm + N
                      left2 = ijklm - N
                      jnext = j + 1
                      jprev = j - 1
                   endif
                   
                   !Dimension 3
                   if (k == N) then
                      right3 = ijklm - (N-1)*N**2
                      left3 = ijklm - N**2
                      knext = 1
                      kprev = N - 1
                   else if (k == 1) then
                      right3 = ijklm + N**2
                      left3 = ijklm + (N-1)*N**2
                      knext = 2
                      kprev = N
                   else
                      right3 = ijklm + N**2
                      left3 = ijklm - N**2
                      knext = k + 1
                      kprev = k - 1
                   endif
                   
                   !Dimension 4
                   if (l == N) then
                      right4 = ijklm - (N-1)*N**3
                      left4 = ijklm - N**3
                      lnext = 1
                      lprev = N - 1
                   else if (l == 1) then
                      right4 = ijklm + N**3
                      left4 = ijklm + (N-1)*N**3
                      lnext = 2
                      lprev = N
                   else
                      right4 = ijklm + N**3
                      left4 = ijklm - N**3
                      lnext = l + 1
                      lprev = l - 1
                   endif
                   
                   !Dimension 5
                   if (m == N) then
                      right5 = ijklm - (N-1)*N**4
                      left5 = ijklm - N**4
                      mnext = 1
                      mprev = N - 1
                   else if (m == 1) then
                      right5 = ijklm + N**4
                      left5 = ijklm + (N-1)*N**4
                      mnext = 2
                      mprev = N
                   else
                      right5 = ijklm + N**4
                      left5 = ijklm - N**4
                      mnext = m + 1
                      mprev = m - 1
                   endif

                   !Chech the neighbours for broken fibres:
                   
                   !Right in D1
                   if (fibres(inext,j,k,l,m) == 0) then
                      hoshk(hoshk(right1,1),3) = hoshk(hoshk(right1,1),3) + 1
                      holes(i,j,k,l,m,1) = hoshk(right1,1)
                      counter = counter + 1
                   endif
                   
                   !Left in D1
                   already_found = .false.
                   if (fibres(iprev,j,k,l,m) == 0) then
                      if (counter == 0) then
                         hoshk(hoshk(left1,1),3) = hoshk(hoshk(left1,1),3) + 1
                         holes(i,j,k,l,m,1) = hoshk(left1,1)
                         counter = counter + 1
                      else
                         do a = 1,counter
                            if (hoshk(left1,1) == holes(i,j,k,l,m,a)) then
                               already_found = .true.
                            endif
                         enddo
                         if (.not. already_found) then
                            hoshk(hoshk(left1,1),3) = hoshk(hoshk(left1,1),3) + 1
                            holes(i,j,k,l,m,counter+1) = hoshk(left1,1)
                            counter = counter + 1
                         endif
                      endif
                   endif
                   
                   !Right in D2
                   already_found = .false.
                   if (fibres(i,jnext,k,l,m) == 0) then
                      if (counter == 0) then
                         hoshk(hoshk(right2,1),3) = hoshk(hoshk(right2,1),3) + 1
                         holes(i,j,k,l,m,1) = hoshk(right2,1)
                         counter = counter + 1
                      else
                         do a = 1,counter
                            if (hoshk(right2,1) == holes(i,j,k,l,m,a)) then
                               already_found = .true.
                            endif
                         enddo
                         if (.not. already_found) then
                            hoshk(hoshk(right2,1),3) = hoshk(hoshk(right2,1),3) + 1
                            holes(i,j,k,l,m,counter+1) = hoshk(right2,1)
                            counter = counter + 1
                         endif
                      endif
                   endif
                   
                   !Left in D2
                   already_found = .false.
                   if (fibres(i,jprev,k,l,m) == 0) then
                      if (counter == 0) then
                         hoshk(hoshk(left2,1),3) = hoshk(hoshk(left2,1),3) + 1
                         holes(i,j,k,l,m,1) = hoshk(left2,1)
                         counter = counter + 1
                      else
                         do a = 1,counter
                            if (hoshk(left2,1) == holes(i,j,k,l,m,a)) then
                               already_found = .true.
                            endif
                         enddo
                         if (.not. already_found) then
                            hoshk(hoshk(left2,1),3) = hoshk(hoshk(left2,1),3) + 1
                            holes(i,j,k,l,m,counter+1) = hoshk(left2,1)
                            counter = counter + 1
                         endif
                      endif
                   endif
                   
                   !Right in D3
                   already_found = .false.
                   if (fibres(i,j,knext,l,m) == 0) then
                      if (counter == 0) then
                         hoshk(hoshk(right3,1),3) = hoshk(hoshk(right3,1),3) + 1
                         holes(i,j,k,l,m,1) = hoshk(right3,1)
                         counter = counter + 1
                      else
                         do a = 1,counter
                            if (hoshk(right3,1) == holes(i,j,k,l,m,a)) then
                               already_found = .true.
                            endif
                         enddo
                         if (.not. already_found) then
                            hoshk(hoshk(right3,1),3) = hoshk(hoshk(right3,1),3) + 1
                            holes(i,j,k,l,m,counter+1) = hoshk(right3,1)
                            counter = counter + 1
                         endif
                      endif
                   endif
                   
                   !Left in D3
                   already_found = .false.
                   if (fibres(i,j,kprev,l,m) == 0) then
                      if (counter == 0) then
                         hoshk(hoshk(left3,1),3) = hoshk(hoshk(left3,1),3) + 1
                         holes(i,j,k,l,m,1) = hoshk(left3,1)
                         counter = counter + 1
                      else
                         do a = 1,counter
                            if (hoshk(left3,1) == holes(i,j,k,l,m,a)) then
                               already_found = .true.
                            endif
                         enddo
                         if (.not. already_found) then
                            hoshk(hoshk(left3,1),3) = hoshk(hoshk(left3,1),3) + 1
                            holes(i,j,k,l,m,counter+1) = hoshk(left3,1)
                            counter = counter + 1
                         endif
                      endif
                   endif
                   
                   !Right in D4
                   already_found = .false.
                   if (fibres(i,j,k,lnext,m) == 0) then
                      if (counter == 0) then
                         hoshk(hoshk(right4,1),3) = hoshk(hoshk(right4,1),3) + 1
                         holes(i,j,k,l,m,1) = hoshk(right4,1)
                         counter = counter + 1
                      else
                         do a = 1,counter
                            if (hoshk(right4,1) == holes(i,j,k,l,m,a)) then
                               already_found = .true.
                            endif
                         enddo
                         if (.not. already_found) then
                            hoshk(hoshk(right4,1),3) = hoshk(hoshk(right4,1),3) + 1
                            holes(i,j,k,l,m,counter+1) = hoshk(right4,1)
                            counter = counter + 1
                         endif
                      endif
                   endif
                   
                   !Left in D4
                   already_found = .false.
                   if (fibres(i,j,k,lprev,m) == 0) then
                      if (counter == 0) then
                         hoshk(hoshk(left4,1),3) = hoshk(hoshk(left4,1),3) + 1
                         holes(i,j,k,l,m,1) = hoshk(left4,1)
                         counter = counter + 1
                      else
                         do a = 1,counter
                            if (hoshk(left4,1) == holes(i,j,k,l,m,a)) then
                               already_found = .true.
                            endif
                         enddo
                         if (.not. already_found) then
                            hoshk(hoshk(left4,1),3) = hoshk(hoshk(left4,1),3) + 1
                            holes(i,j,k,l,m,counter+1) = hoshk(left4,1)
                            counter = counter + 1
                         endif
                      endif
                   endif
                   
                   !Right in D5
                   already_found = .false.
                   if (fibres(i,j,k,l,mnext) == 0) then
                      if (counter == 0) then
                         hoshk(hoshk(right5,1),3) = hoshk(hoshk(right5,1),3) + 1
                         holes(i,j,k,l,m,1) = hoshk(right5,1)
                         counter = counter + 1
                      else
                         do a = 1,counter
                            if (hoshk(right5,1) == holes(i,j,k,l,m,a)) then
                               already_found = .true.
                            endif
                         enddo
                         if (.not. already_found) then
                            hoshk(hoshk(right5,1),3) = hoshk(hoshk(right5,1),3) + 1
                            holes(i,j,k,l,m,counter+1) = hoshk(right5,1)
                            counter = counter + 1
                         endif
                      endif
                   endif
                   
                   !Left in D5
                   already_found = .false.
                   if (fibres(i,j,k,l,mprev) == 0) then
                      if (counter == 0) then
                         hoshk(hoshk(left5,1),3) = hoshk(hoshk(left5,1),3) + 1
                         holes(i,j,k,l,m,1) = hoshk(left5,1)
                         counter = counter + 1
                      else
                         do a = 1,counter
                            if (hoshk(left5,1) == holes(i,j,k,l,m,a)) then
                               already_found = .true.
                            endif
                         enddo
                         if (.not. already_found) then
                            hoshk(hoshk(left5,1),3) = hoshk(hoshk(left5,1),3) + 1
                            holes(i,j,k,l,m,counter+1) = hoshk(left5,1)
                            counter = counter + 1
                         endif
                      endif
                   endif

                endif
             enddo
          enddo
       enddo
    enddo
    enddo

  end subroutine hoshen_kopelman
end module basic_functions
